/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __DSA_DSM_API_WRAPPER_H__
#define __DSA_DSM_API_WRAPPER_H__

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <digitalSTROM/dsm-api-v2/dsm-api.h>

#include <boost/thread/mutex.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <string>
#include <list>
#include <vector>

#define DSM_API_MAX_DEVICES 63

/// \file dsmapi-wrapper.h
namespace dSA
{

enum
{
    UPSTREAM_FREQUENCY_15_20    = 0x00, // 15 kHz / 20 kHz
    UPSTREAM_FREQUENCY_12_20    = 0x01, // 12 kHz / 20 kHz
    UPSTREAM_FREQUENCY_10_15    = 0x02, // 10 kHz / 15 kHz
    UPSTREAM_FREQUENCY_10_12    = 0x03  // 10 kHz / 12 kHz
};

typedef struct
{
    uint16_t version; /**< Version of the configuration structure.*/
    uint16_t dSMC_DAC1a; /**< Offset value defining the start of the first downstream pulse.*/
    uint16_t dSMC_DAC2a; /**< Offset value defining the start of the second downstream pulse.*/
    uint16_t dSMC_OFFSET; /**< Current offset for the dSMC.*/
    uint16_t dSMC_clock; /**< Clock setting for the dSMC.*/
    uint16_t mains_offset; /**< Mains voltage offset.*/
    uint8_t dSMC_BITD; /**< Register BITD im dSMC.*/
    uint8_t dSMC_BLANK1; /**< Register BLANK1 im dSMC.*/
    uint8_t dSMC_BLANK2; /**< Register BLANK2 im dSMC.*/
    uint8_t dSMC_BCRS; /**< Register BCRS im dSMC.*/
    uint8_t dSMC_PRGR; /**< Register PRGR im dSMC.*/
    uint8_t dSMC_TESTMUX; /**< Register TESTMUX im dSMC.*/
    uint16_t upstream_start_gaps[2]; /**< Number of samples pause before start of front sampling window. Due to offset of POL signal two values for positive and negative halfwaves.*/
    uint16_t upstream_between_gap; /**< Number of samples pause between the front and back sampling windows.*/
    uint8_t upstream_carrier_threshold; /**< Carrier detection threshold.*/
    uint8_t downstream_mode; /**< Downstream modulation mode (#downstream_mode_enum).*/
    uint8_t downstream_single_bit_shape; /**< Shape of single bits in downstream_mode_[2|3]bit.*/
    uint8_t upstream_noise_floor; /**< the noise average used for carrier detection will never drop below that value expressed as a power of two.*/
    uint32_t power_calibration_high; /**< Calibration value for the high power calculation. round(real_value / measured_value * 2^18).*/
    uint32_t power_calibration_low; /**< Calibration value for the low power calculation. round(real_value / measured_value * 2^21)*/
    uint16_t power_voltage_offset; /**< Offset of the voltage measurement.*/
    uint16_t power_current_offset_high; /**< Offset of the high current measurement.*/
    uint16_t power_current_offset_low; /**< Offset of the low current measurement.*/
    uint16_t temperature_offset; /**< Offset of the temperature measurement.*/
    uint16_t downstream_debug; /**< Iff != 0: Continuously send content downstream.*/
    uint16_t trigger_MODI; /**< Trigger level for the MODI check.*/
} DsmDspConfigData_t;

#define BUSMEMBER_IS_DSM(type) ((type == BusMember_dSM11) || \
                                (type == BusMember_dSM12) || \
                                (type == BusMember_vDC))
class dSMApiListener
{
public:
    virtual ~dSMApiListener() {}

    /// dS485-Bus-Level Callbacks
    virtual void busStateCallback(bus_state_t state) = 0;
    virtual void busChangeCallback(dsuid_t *id, int flag) = 0;

    /// Device-Level Callbacks
    virtual void deviceActiveCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                     dsuid_t dest_dsm_dsuid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     dsuid_t dsuid) = 0;
    virtual void deviceInactiveCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                       dsuid_t dest_dsm_dsuid, uint16_t device_id,
                                       uint16_t zone_id, uint16_t vendor_id,
                                       dsuid_t dsuid) = 0;
    virtual void devicePingCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
            dsuid_t dest_dsm_dsuid, uint16_t device_id, uint8_t reponse_flag) = 0;
    virtual void deviceStatisticsCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint16_t value) = 0;
    virtual void deviceActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type) = 0;
    virtual void deviceLocalActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t state) = 0;
    virtual void deviceConfigCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value) = 0;
    virtual void deviceTestTransmissionQualityCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality) = 0;
    virtual void deviceRemovedByDsuidCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid) = 0;
    virtual void deviceRemovedByIdCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id) = 0;

    /// Communication Callbacks
    virtual void commUpstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags) = 0;
    virtual void commUpstreamLongCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags) = 0;
    virtual void commDownstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command) = 0;
    virtual void commDownstreamLong0Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command) = 0;
    virtual void commDownstreamLong8Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint8_t data) = 0;
    virtual void commDownstreamLong16Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint16_t data) = 0;

    // dSMeter Callbacks
    virtual void dsmStatusCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
            dsuid_t dest_dsm_dsuid, uint8_t hardware, uint8_t communicationdsp,
            uint8_t temperature, uint8_t configuration, uint8_t ds485bus,
            uint8_t powerline) = 0;
    virtual void dsmErrorCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
            dsuid_t dest_dsm_dsuid, uint16_t error, uint32_t data) = 0;
    virtual void dsmRebootRequestCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
            dsuid_t dest_dsm_dsuid) = 0;
};

class dSMApiSelectiveListener : public dSMApiListener
{
public:
    typedef enum
    {
        DsmApiBusState,
        DsmApiBusChange,
        DsmApiDeviceActive,
        DsmApiDeviceInactive,
        DsmApiDevicePing,
        DsmApiDeviceStatistics,
        DsmApiDeviceAction,
        DsmApiDeviceLocalAction,
        DsmApiDeviceConfig,
        DsmApiDeviceTestTransmissionQuality,
        DsmApiDeviceRemovedByDsuid,
        DsmApiDeviceRemovedById,
        DsmApiCommUpstreamShort,
        DsmApiCommUpstreamLong,
        DsmApiCommDownstreamShort,
        DsmApiCommDownstreamLong0,
        DsmApiCommDownstreamLong8,
        DsmApiCommDownstreamLong16,
        DsmApiDsmState,
        DsmApiDsmError,
        DsmApiDsmRebootRequest,

        CB_MAX
    } cb_selection;

    class CBData
    {
    public:
        CBData(cb_selection type);
        virtual ~CBData() {}
        cb_selection getType() const;
        time_t getTimestamp() const;
    private:
        CBData();
        cb_selection m_data_type;
        time_t m_timestamp;
    };

    class CBCommonData : public CBData
    {
    public:
        CBCommonData(cb_selection type, uint8_t error_code,
               dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid);
        uint8_t getErrorCode() const;
        dsuid_t getDSMSourceDSUID() const;
        dsuid_t getDSMDestinationDSUID() const;
        virtual ~CBCommonData() {}
    private:
        uint8_t m_error_code;
        dsuid_t m_src_dsm_dsuid;
        dsuid_t m_dest_dsm_dsuid;
    };

    class BasicDeviceData : public CBCommonData
    {
    public:
        BasicDeviceData(cb_selection type, uint8_t error_code,
                dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id);
        uint16_t getDeviceID() const;
        virtual ~BasicDeviceData() {}
    private:
        uint16_t m_device_id;
    };

    class DeviceData : public BasicDeviceData
    {
    public:
        DeviceData(cb_selection type, uint8_t error_code,
                dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
                uint16_t zone_id, uint16_t vendor_id, dsuid_t dsuid);
        uint16_t getZoneID() const;
        uint16_t getVendorID() const;
        virtual ~DeviceData() {}
    private:
        uint16_t m_zone_id;
        uint16_t m_vendor_id;
    };

    class DeviceActiveData: public DeviceData
    {
    public:
        DeviceActiveData(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint16_t device_id, uint16_t zone_id,
                uint16_t vendor_id, dsuid_t dsuid);
        virtual ~DeviceActiveData() {}
    };

    class DeviceInactiveData: public DeviceData
    {
    public:
        DeviceInactiveData(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint16_t device_id, uint16_t zone_id,
                uint16_t vendor_id, dsuid_t dsuid);
        virtual ~DeviceInactiveData() {}
    };

    class DevicePingData: public BasicDeviceData
    {
    public:
        DevicePingData(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint16_t device_id, uint8_t present);
        uint8_t getPresent() const;
        virtual ~DevicePingData() {}
    private:
        uint8_t m_present;
    };

    class DeviceTestQualityData: public BasicDeviceData
    {
    public:
        DeviceTestQualityData(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint16_t device_id, uint8_t down, uint16_t up);
        uint8_t getDownstream() const;
        uint16_t getUpstream() const;
        virtual ~DeviceTestQualityData() {}
    private:
        uint8_t m_down;
        uint16_t m_up;
    };

    class DeviceStatisticData: public BasicDeviceData
    {
    public:
        DeviceStatisticData(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint16_t device_id,
                uint8_t offset, uint16_t value);
        uint8_t getOffset() const;
        uint16_t getValue() const;
        virtual ~DeviceStatisticData() {}
    private:
        uint8_t m_offset;
        uint16_t m_value;
    };

    class CommData : public BasicDeviceData
    {
    public:
        CommData(cb_selection type, uint8_t error_code, dsuid_t src_dsm_dsuid,
                 dsuid_t dest_dsm_dsuid, uint16_t device_id,
                 uint8_t circuit_id, uint8_t quality,
                 uint8_t flags, uint8_t cflags);
        uint8_t getCircuitID() const;
        uint8_t getQuality() const;
        uint8_t getFlags() const;
        uint8_t getCrosstalkFlags() const;
        virtual ~CommData() {}
    private:
        uint8_t m_circuit_id;
        uint8_t m_quality;
        uint8_t m_flags;
        uint8_t m_cflags;
    };

    class CommUpstreamShortData : public CommData
    {
    public:
        CommUpstreamShortData(uint8_t error_code, dsuid_t src_dsm_dsuid,
            dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type, uint8_t quality,
            uint8_t flags, uint8_t cflags);
        uint8_t isResend() const;
        uint8_t isSensor() const;
        uint8_t getKeyType() const;
        uint8_t getClickType() const;
        virtual ~CommUpstreamShortData() {}
    private:
        uint8_t m_is_resend;
        uint8_t m_is_sensor;
        uint8_t m_key_type;
        uint8_t m_click_type;
    };

    class CommUpstreamLongData : public CommData
    {
    public:
        CommUpstreamLongData(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);
        uint8_t getMainQualifier() const;
        uint8_t getSubQualifier() const;
        uint32_t getData() const;
        virtual ~CommUpstreamLongData() {}
    private:
        uint8_t m_mainqualifier;
        uint8_t m_subqualifier;
        uint32_t m_data;
    };

    class CommDownstreamData : public CBCommonData
    {
    public:
        CommDownstreamData(cb_selection type, uint8_t error_code,
                dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);
        uint8_t getCommand() const;
        virtual ~CommDownstreamData() {}
    private:
        uint8_t m_command;
    };

    class CommDownstreamShortData : public CommDownstreamData
    {
    public:
        CommDownstreamShortData(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint8_t command);
        virtual ~CommDownstreamShortData() {}
    };

    class CommDownstreamLong0Data : public CommDownstreamData
    {
    public:
        CommDownstreamLong0Data(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint8_t command);
        virtual ~CommDownstreamLong0Data() {}
    };

    class CommDownstreamLong8Data : public CommDownstreamData
    {
    public:
        CommDownstreamLong8Data(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint8_t command, uint8_t data);
        uint8_t getData() const;
        virtual ~CommDownstreamLong8Data() {}
    private:
        uint8_t m_data;
    };

    class CommDownstreamLong16Data : public CommDownstreamData
    {
    public:
        CommDownstreamLong16Data(uint8_t error_code, dsuid_t src_dsm_dsuid,
                dsuid_t dest_dsm_dsuid, uint8_t command, uint16_t data);
        uint16_t getData() const;
        virtual ~CommDownstreamLong16Data() {}
    private:
        uint16_t m_data;
    };

    class DsmStatusData : public CBCommonData
    {
    public:
        DsmStatusData(uint8_t error_code,
                dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
                uint8_t hardware, uint8_t communicationdsp,
                uint8_t temperature, uint8_t configuration,
                uint8_t ds485bus, uint8_t powerline);
        uint8_t getHardware() const;
        uint8_t getCommunicationDSP() const;
        uint8_t getTemperature() const;
        uint8_t getConfiguration() const;
        uint8_t getDS485Bus() const;
        uint8_t getPowerline() const;
        virtual ~DsmStatusData() {}
    private:
        uint8_t m_hardware;
        uint8_t m_communicationdsp;
        uint8_t m_temperature;
        uint8_t m_configuration;
        uint8_t m_ds485bus;
        uint8_t m_powerline;
    };

    class DsmErrorData : public CBCommonData
    {
    public:
        DsmErrorData(uint8_t error_code,
                dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
                uint16_t error, uint32_t data);
        uint16_t getError() const;
        uint32_t getData() const;
        virtual ~DsmErrorData() {}
    private:
        uint16_t m_error;
        uint32_t m_data;
    };

    class DsmRebootRequestData : public CBCommonData
    {
    public:
        DsmRebootRequestData(uint8_t error_code,
                dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid);
        virtual ~DsmRebootRequestData() {}
    };

    class BusChangeData : public CBData
    {
    public:
        BusChangeData(dsuid_t dsm_dsuid, int status);
        dsuid_t getDSUID() const;
        int getStatus() const;
        virtual ~BusChangeData() {}
    private:
        dsuid_t m_dsuid;
        int m_status;
    };

    dSMApiSelectiveListener();
    virtual ~dSMApiSelectiveListener();
    void select(dSMApiSelectiveListener::cb_selection cb, bool on);

    virtual void busStateCallback(bus_state_t state);
    virtual void busChangeCallback(dsuid_t *id, int flag);

    /// Device-Level Callbacks
    virtual void deviceActiveCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                     dsuid_t dest_dsm_dsuid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     dsuid_t dsuid);
    virtual void deviceInactiveCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                       dsuid_t dest_dsm_dsuid, uint16_t device_id,
                                       uint16_t zone_id, uint16_t vendor_id,
                                       dsuid_t dsuid);
    virtual void devicePingCallback(uint8_t error_code, dsuid_t src_dsm_dsuid,
            dsuid_t dest_dsm_dsuid, uint16_t device_id, uint8_t reponse_flag);
    virtual void deviceStatisticsCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint16_t value);
    virtual void deviceActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type);
    virtual void deviceLocalActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);
    virtual void deviceConfigCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    virtual void deviceTestTransmissionQualityCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);
    virtual void deviceRemovedByDsuidCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid);
    virtual void deviceRemovedByIdCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id);

    /// Communication Callbacks
    virtual void commUpstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type, uint8_t quality,
            uint8_t flags, uint8_t cflags);
    virtual void commUpstreamLongCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);
    virtual void commDownstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);
    virtual void commDownstreamLong0Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);
    virtual void commDownstreamLong8Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint8_t data);
    virtual void commDownstreamLong16Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint16_t data);

    // dSMeter Callbacks
    virtual void dsmStatusCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint8_t hardware, uint8_t communicationdsp,
            uint8_t temperature, uint8_t configuration,
            uint8_t ds485bus, uint8_t powerline);
    virtual void dsmErrorCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint16_t error, uint32_t data);
    virtual void dsmRebootRequestCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid);

    virtual void callback(boost::shared_ptr<CBData> callback_data) = 0;
private:
    bool checkCB(cb_selection sel);
    std::vector<bool> m_selection;
    boost::recursive_mutex m_sel_access_mutex;
};

/// This class provides a threadsafe access to the dSM API library.
///
/// Each new instance of the class will open a new connection to the dS485
/// daemon. However each instance is threadsafe and can be used from various
/// threads. Currently it will only wrap the functions that we are interested
/// in, more calls will be added on demand.
///
/// Instead of return values the wrapped calls to the lib will throw exceptions.
class dSMApiWrapper
{
public:
    dSMApiWrapper(std::string ds485_daemon_url);
    ~dSMApiWrapper();

    void addListener(dSMApiListener* listener);
    void removeListener(dSMApiListener* listener);

    void getOwnDSUID(dsuid_t* dsuid);
    bus_state_t getBusState();
    int getBusMembersPerType(dsuid_t *device_list, uint16_t types);
    int getBusMemberType(dsuid_t busMember);

    /// dSMeter Infos

    void dSMInfo(dsuid_t dsm_dsuid, uint32_t *hw_version,
                 uint32_t *arm_sw_version, uint32_t *dsp_sw_version,
                 uint16_t *api_version, dsid_t *dsid, uint8_t *name);
    void dSMState(dsuid_t dsm_dsuid, uint8_t *hardware, uint8_t *communicationdsp,
                  uint8_t *temperature, uint8_t *configuration,
                  uint8_t *ds485bus, uint8_t *powerline);
    void getPropertyFlags(dsuid_t dsm_dsuid, uint8_t *flags);
    void setPropertyFlags(dsuid_t dsm_dsuid, uint8_t flags);
    void getPropertyTemperatureLimit(dsuid_t dsm_dsuid, uint8_t *warnlevel, uint8_t *errorlevel);
    void setPropertyTemperatureLimit(dsuid_t dsm_dsuid, uint8_t warnlevel, uint8_t errorlevel);
    void getDSMtemperature(dsuid_t dsm_dsuid, uint16_t *temperature);
    void getApartmentId(dsuid_t dsm_dsuid, uint8_t *id);
    void setApartmentId(dsuid_t dsm_dsuid, uint8_t id);
    void getUpstreamFrequency(dsuid_t dsm_dsuid, uint8_t *freq);
    void setUpstreamFrequency(dsuid_t dsm_dsuid, uint8_t freq);
    void getConfigurationId(dsuid_t dsm_dsuid, uint32_t *hash,
                            uint32_t *configurationcount);
    void getStatisticUpstream(dsuid_t dsm_dsuid, uint8_t index,
            uint8_t *length, uint8_t *statistics);
    void resetStatisticUpstream(dsuid_t dsm_dsuid);

    /// Infrastructure

    void getZoneCount(dsuid_t dsm_dsuid, uint8_t *count);
    void getZoneInfo(dsuid_t dsm_dsuid, uint8_t zone_index, uint16_t *zone_id,
                      uint8_t *virtual_zone_id, uint8_t *number_of_groups,
                      uint8_t *name);
    void getActiveDevicesCountInZone(dsuid_t dsm_dsuid, uint16_t zone_id,
                                     uint16_t *number_of_devices);
    void getInactiveDevicesCountInZone(dsuid_t dsm_dsuid, uint16_t zone_id,
                                       uint16_t *number_of_devices);
    void getDevicesCountInZone(dsuid_t dsm_dsuid, uint16_t zone_id,
                               uint16_t *number_of_devices);
    void getActiveDeviceInfoInZone(dsuid_t dsm_dsuid, uint16_t req_zone_id,
                             uint16_t device_index, uint16_t *device_id,
                             uint16_t *vendor_id, uint16_t *product_id,
                             uint16_t *function_id, uint16_t *version,
                             uint16_t *zone_id, uint8_t *state,
                             uint8_t *locked,uint8_t *output_mode,
                             uint8_t *groups, uint8_t *name,
                             dsuid_t *dsuid, uint8_t *lt_mode);
    void getDeviceInfoById(dsuid_t dsm_dsuid, uint16_t req_device_id,
                       uint16_t *device_id, uint16_t *vendor_id,
                       uint16_t *product_id, uint16_t *function_id,
                       uint16_t *version, uint16_t *zone_id, uint8_t *state,
                       uint8_t *locked, uint8_t *output_mode, uint8_t *groups,
                       uint8_t *name, dsuid_t *dsuid, uint8_t *lt_mode);
    void getDeviceInfoByIndex(dsuid_t dsm_dsuid, uint16_t req_zone_id,
                       uint16_t req_index,
                       uint16_t *device_id, uint16_t *vendor_id,
                       uint16_t *product_id, uint16_t *function_id,
                       uint16_t *version, uint16_t *zone_id, uint8_t *state,
                       uint8_t *locked, uint8_t *output_mode, uint8_t *groups,
                       uint8_t *name, dsuid_t *dsuid, uint8_t *lt_mode);

    // dSMeter Actions
    void restartRegistration(dsuid_t dsm_dsuid);
    void restartMeter(dsuid_t dsm_dsuid);
    void setDisplayMode(dsuid_t dsm_dsuid, uint8_t flags);
    void getDisplayMode(dsuid_t dsm_dsuid, uint8_t* flags);
    void setCommunicationLogging(dsuid_t dsm_dsuid, uint8_t state);
    void getCommunicationLogging(dsuid_t dsm_dsuid, uint8_t* state);
    void getDspConfig(dsuid_t dsm_dsuid, DsmDspConfigData_t* config);
    void writeDspConfig(dsuid_t dsm_dsuid, const DsmDspConfigData_t &config);
    void restartDsp(dsuid_t dsm_dsuid);

    // dSMeter device presence check
    void pingAutomatic(dsuid_t dsm_dsuid, uint16_t auto_interval_time);
    void pingOneTimeAutomatic(dsuid_t dsm_dsuid, uint16_t interval_time,
                              uint16_t auto_interval_time);
    void pingOneTimeOff(dsuid_t dsm_dsuid, uint16_t interval_time);
    void pingOff(dsuid_t dsm_dsuid);
    void getPingIntervals(dsuid_t dsm_dsuid, uint16_t *interval_auto,
                          uint16_t *interval_device);
    void setPingIntervals(dsuid_t dsm_dsuid, uint16_t interval_auto,
                          uint16_t interval_device);

    /// Device Actions
    void blinkDevice(dsuid_t dsm_dsuid, uint16_t device_id);
    void setConfigurationValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    void getConfigurationValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t *value);
    void getConfigurationValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint16_t *value);
    void getStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint8_t *value);
    void getStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint16_t *value);
    void getStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint32_t *value);
    void reqStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t offset);
    void getSensorValue(dsuid_t dsm_dsuid, uint16_t device_id,
            uint8_t index, uint16_t *value);
    void testDeviceTransmissionQuality(dsuid_t dsm_dsuid, uint16_t device_id,
                                        uint8_t *downstreamQ,
                                        uint16_t *upstreamQ);
    void reqDeviceTransmissionQuality(dsuid_t dsm_dsuid, uint16_t device_id);
    void reqDevicePresenceTest(dsuid_t dsm_dsuid, uint16_t device_id);

    // virtual device conection
    void protobufMessageRequest(dsuid_t dsm_dsuid, uint16_t request_size,
                                uint8_t *request, uint16_t *response_size,
                                uint8_t *response);

private:
    boost::mutex m_libdsm_mutex;
    boost::mutex m_listener_mutex;
    std::string m_daemon_url;
    DsmApiHandle_t m_dsm_api_handle;
    std::list<dSMApiListener*> m_callback_listeners;

    void checkRetcode(int ret);

    static void busStateCallback(void *arg, bus_state_t state);
    void notifyBusState(bus_state_t state);

    static void busChangeCallback(void *arg, dsuid_t *id, int flag);
    void notifyBusChange(dsuid_t *id, int flag);

    static void deviceActiveCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint16_t product_id,
            uint16_t function_id, uint16_t version, dsuid_t dsuid);
    void notifyDeviceActive(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, dsuid_t dsuid);

    static void deviceInactiveCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint16_t product_id,
            uint16_t function_id, uint16_t version, dsuid_t dsuid);
    void notifyDeviceInactive(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, dsuid_t dsuid);

    static void devicePingCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t present);
    void notifyDevicePing(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t present);

    static void deviceStatisticsCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint16_t value);
    void notifyDeviceStatistics(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint16_t value);

    static void deviceActionCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type);
    void notifyDeviceActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type);

    static void deviceLocalActionCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);
    void notifyDeviceLocalActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);

    static void deviceConfigCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    void notifyDeviceConfigCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);

    static void deviceTestTransmissionQualityCallback(uint8_t error_code,
            void *arg, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint16_t device_id, uint8_t downstream_quality,
            uint16_t upstream_quality);
    void notifyDeviceTestTransmissionQualityCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);

    static void deviceRemovedByDsuidCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid);
    void notifyDeviceRemovedByDsuidCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid);

    static void deviceRemovedByIdCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id);
    void notifyDeviceRemovedByIdCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id);

    static void commUpstreamShortCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags);
    void notifyCommUpstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags);

    static void commUpstreamLongCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);
    void notifyCommUpstreamLongCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);

    static void commDownstreamShortCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);
    void notifyCommDownstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);

    static void commDownstreamLong0Callback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);
    void notifyCommDownstreamLong0Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command);

    static void commDownstreamLong8Callback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint8_t data);
    void notifyCommDownstreamLong8Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint8_t data);

    static void commDownstreamLong16Callback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint16_t data);
    void notifyCommDownstreamLong16Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint16_t data);

    static void dsmStatusCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint8_t hardware, uint8_t communicationdsp,
            uint8_t temperature, uint8_t configuration, uint8_t ds485bus,
            uint8_t powerline);
    void notifyDsmStatusCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint8_t hardware, uint8_t communicationdsp,
            uint8_t temperature, uint8_t configuration, uint8_t ds485bus,
            uint8_t powerline);

    static void dsmErrorCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint16_t error, uint32_t data);
    void notifyDsmErrorCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
            uint16_t error, uint32_t data);

    static void dsmRebootRequestCallback(uint8_t error_code, void *arg,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid);
    void notifyDsmRebootRequestCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid);
};

}; // namespace

#endif//__DSA_DSM_API_WRAPPER_H__

